from typing import List, Set, Tuple, Optional, Type, TypeVar, Generic, Any, Union
import dgl
import torch

from .batch import Batchable, Batch, BatchBase, TypeTree, make_batch, make_batch_td, BatchTD
from .type_data import ClassTD, TypeData, ShapeVar

class Batchable(Batchable):
    pass

N = TypeVar('N', bound=Batchable)
E = TypeVar('E', bound=Optional[Batchable])
class Graph(Generic[N, E], Batchable):
    """ Wrapper around dgl graph allowing easier access to data """

    dgl_graph: dgl.graph
    node_type_tree: TypeTree
    edge_type_tree: Optional[TypeTree]

    @staticmethod
    def get_batch_type():
        return GraphBatch

    @staticmethod
    def get_batch_td_type():
        return GraphBatchTD

    def __init__(self, nodes: List[N],
                 edges: List[Tuple[int, int]],
                 edata: Optional[List[E]] = None,
                 directed: bool = False):
        """ If directed is false, both permutations of the edges
        will be added. """

        # just like 
        if nodes is None: return

        src_list = []
        dst_list = []
        new_edata = None if edata is None else []
        for i, (n1, n2) in enumerate(edges):
            src_list.append(n1)
            dst_list.append(n2)
            if new_edata is not None:
                new_edata.append(edata[i])
            if not directed:
                src_list.append(n2)
                dst_list.append(n1)
                if new_edata is not None:
                    new_edata.append(edata[i])
        
        self.dgl_graph = dgl.graph((torch.tensor(src_list), torch.tensor(dst_list)), num_nodes=len(nodes), idtype=torch.int32, device='cpu')

        node_batch = Batch(nodes)
        self.node_type_tree = node_batch.type_tree
        for key, val in node_batch.store.items():
            self.dgl_graph.ndata[key] = val

        if new_edata is not None:
            edge_batch = Batch(new_edata)
            self.edge_type_tree = edge_batch.type_tree
            for key, val in edge_batch.store.items():
                self.dgl_graph.edata[key] = val
        else:
            self.edge_type_tree = None

    @property
    def ndata(self) -> Batch[N]:
        ret = Batch(None)
        ret.batch_size = self.dgl_graph.number_of_nodes()
        ret.type_tree = self.node_type_tree
        ret.store = self.dgl_graph.ndata
        return ret

    @property
    def edata(self) -> Batch[E]:
        if self.edge_type_tree is None:
            raise AttributeError
        ret = Batch(None)
        ret.batch_size = self.dgl_graph.number_of_edges()
        ret.type_tree = self.edge_type_tree
        ret.store = self.dgl_graph.edata
        return ret

    @property
    def edges(self) -> List[Tuple[int, int]]:
        ret = []
        for src, dst in zip(*self.dgl_graph.edges()):
            ret.append((src, dst))
        return ret

    def to(self, device):
        ret = Graph.__new__(GraphBatch)
        ret.node_type_tree = self.node_type_tree
        ret.edge_type_tree = self.edge_type_tree
        ret.dgl_batch = self.dgl_graph.to(device)
        return ret

G = TypeVar('G', bound=Graph)
class GraphBatch(BatchBase[G]):

    dgl_batch: dgl.batch
    graph_type: Type[Graph]
    node_type_tree: TypeTree
    edge_type_tree: Optional[TypeTree]
    
    def __init__(self, items: List[Graph[N, E]]):
        assert len(items) > 0
        first = items[0]
        self.graph_type = type(first)
        self.node_type_tree = first.node_type_tree
        self.edge_type_tree = first.edge_type_tree
        self.dgl_batch = dgl.batch([ item.dgl_graph for item in items ])

    def __len__(self) -> int:
        return self.dgl_batch.batch_size

    def __getitem__(self, index: int) -> Graph[N, E]:
        if index >= len(self):
            raise IndexError
        ret = self.graph_type.__new__(self.graph_type)
        ret.dgl_graph = dgl.unbatch(self.dgl_batch)[index]
        ret.node_type_tree = self.node_type_tree
        ret.edge_type_tree = self.edge_type_tree
        return ret

    @property
    def ndata(self) -> Batch[N]:
        ret = Batch(None)
        ret.batch_size = self.dgl_batch.number_of_nodes()
        ret.type_tree = self.node_type_tree
        ret.store = self.dgl_batch.ndata
        return ret

    @property
    def edata(self) -> Batch[E]:
        if self.edge_type_tree is None:
            raise AttributeError
        ret = Batch(None)
        ret.batch_size = self.dgl_batch.number_of_edges()
        ret.type_tree = self.edge_type_tree
        ret.store = self.dgl_batch.edata
        return ret

    @property
    def edges(self) -> List[Tuple[int, int]]:
        ret = []
        for src, dst in zip(*self.dgl_batch.edges()):
            ret.append((src, dst))
        return ret

    def to(self, device):
        from multiprocessing import current_process
        ret = GraphBatch.__new__(GraphBatch)
        ret.graph_type = self.graph_type
        ret.node_type_tree = self.node_type_tree
        ret.edge_type_tree = self.edge_type_tree
        ret.dgl_batch = self.dgl_batch.to(device)
        return ret

class GraphTD(ClassTD):

    def __init__(self, runtime_type: Type[Graph],
                 node_td: TypeData,
                 edge_td: TypeData,
                 node_shapevar: Union[ShapeVar, int] = ShapeVar('N'),
                 edge_shapevar: Union[ShapeVar, int] = ShapeVar('E')):
        ndata = BatchTD(node_td, node_shapevar)
        edata = BatchTD(edge_td, node_shapevar)
        super().__init__(runtime_type, ndata=ndata, edata=edata)

class GraphBatchTD(ClassTD):

    batch_size: Union[ShapeVar, int]

    def __init__(self, graph_td, 
                 batch_size: Union[ShapeVar, int] = ShapeVar("B")):
        subtypes = graph_td.subtypes
        runtime_type = GraphBatch[graph_td.runtime_type]
        super().__init__(runtime_type, **subtypes)
        self.batch_size = batch_size

    def __getattr__(self, key: str) -> TypeData:
        if key == "batch_size":
            return self.__dict__[key]
        else:
            return super().__getattr__(key)
        
    def __setattr__(self, key: str, val: TypeData):
        if key == "batch_size":
            self.__dict__[key] = val
        else:
            super().__setattr__(key, val)
    
if __name__ == "__main__":

    class SubNTest(Batchable):
        t1: torch.Tensor

    class NTest(Batchable):
        t1: SubNTest
        t2: torch.Tensor
    
    class Edata(Batchable):
        et1: torch.Tensor

    class TwoGraphs(Batchable):
        g1: Graph
        g2: Graph
    
    nodes = [ NTest(SubNTest(torch.tensor([1,0,0])), torch.tensor([0,1,0])) for n in range(10) ]
    edges = [(0,1), (1,2)]
    edata = [ Edata(torch.tensor([0,0,1])) for e in edges ]
    graph = Graph(nodes, edges, edata)
    print(graph.ndata.t1.t1)
    graph.dgl_graph
    
    batch = make_batch([graph, graph])
    for g in batch:
        print(g.edata.et1)

    tg = TwoGraphs(graph, graph)
    print(make_batch([tg, tg, tg]).g1.ndata.t2)



from typing import TypeVar, Generic, Type, Any, Dict, List, Union, Sequence, Optional

from copy import copy
import torch
import torch.utils.data
from torch.utils.data.dataloader import default_collate
from dataclasses import dataclass

from .type_data import TypeData, ClassTD, TensorTD, ShapeVar
from .meta_utils import classclass, get_type_name, default_init

class Batchable:

    def __init__(self, *args, **kwargs):
        """ default constructor that gets inherited """
        # todo: add error messages
        seen = set()
        assert len(args) <= len(self.__annotations__)
        for name, arg in zip(self.__annotations__.keys(), args):
            setattr(self, name, arg)
            seen.add(name)
            
        for key, val in kwargs.items():
            assert key not in seen
            assert key in self.__annotations__
            setattr(self, key, val)
            
    @staticmethod
    def get_batch_type():
        """ override this method if you want a custom batch type
        for your batchable class """
        raise NotImplementedError()

    @staticmethod
    def get_batch_td_type():
        """ If you're using a custom batch type, you'll also need a custom
        BatchTD type """
        raise NotImplementedError()

@dataclass
class TypeTree:
    type: Type
    subtypes: Dict[str, "TypeTree"]
            
T = TypeVar('T')

class BatchBase(Generic[T]):
    pass

class Batch(BatchBase[T]):

    type_tree: TypeTree
    batch_size: int
    store: Dict[str, Any]

    def __init__(self,
                 items: Optional[Union[List[T], Type]],
                 **kwargs):
        # items should only be None when we're making a blank batch
        # and want to manually mess with things (e.g. in get and set attr)
        if items is None: return
        if isinstance(items, type):
            self.init_from_type(items, **kwargs)
        else:
            self.init_from_list(items)

    def init_from_list(self, items: List[T]):
        assert len(items) > 0
        template = items[0]

        self.type_tree = TypeTree(type(template), {})
        self.batch_size = len(items)
        self.store = {}

        if isinstance(template, Batch):
            # recursion out the wazoo
            raise NotImplementedError
        else:
            for key, val in template.__dict__.items():
                self.type_tree.subtypes[key] = TypeTree(type(val), {})
                attribs = [ getattr(item, key) for item in items ]
                collated = collate(attribs)
                if isinstance(collated, Batch):
                    self.type_tree.subtypes[key] = collated.type_tree
                    for key2, val2 in collated.store.items():
                        full_key = key + "/" + key2
                        self.store[full_key] = val2
                else:
                    self.store[key] = collate(attribs)
                
    def init_from_type(self, batch_type: Type, **kwargs):
        self.type_tree = TypeTree(batch_type, batch_type.__annotations__)
        template_key = next(iter(self.type_tree.subtypes.keys()))
        self.batch_size = len(kwargs[template_key])
        self.store = {}

        for key in self.type_tree.subtypes.keys():
            val = kwargs[key]
            assert len(val) == self.batch_size
            setattr(self, key, val)
            

    def __len__(self) -> int:
        return self.batch_size

    def __getattr__(self, key: str) -> Any:
        if key in [ 'type_tree', 'batch_size', 'store' ]:
            raise AttributeError
        if key in self.store:
            return self.store[key]
        else:
            if key in self.type_tree.subtypes:
                sub_batch = Batch(None)
                sub_batch.type_tree = self.type_tree.subtypes[key]
                sub_batch.batch_size = len(self)
                sub_batch.store = {}
                for key2, val in self.store.items():
                    if key2.startswith(key):
                        first, *rest = key2.split("/")
                        new_key = "/".join(rest)
                        sub_batch.store[new_key] = val
                return sub_batch
            else:
                raise AttributeError
        
    def __setattr__(self, key: str, val: Any):
        # for the sake of the subclasses
        if type(self) != Batch:
            return super().__setattr__(key, val)
        if key in [ 'type_tree', 'batch_size', 'store' ]:
            return super().__setattr__(key, val)
        self.store[key] = val

    def __getitem__(self, index: int) -> T:
        if index >= self.batch_size:
            raise IndexError
        ret = self.type_tree.type.__new__(self.type_tree.type)
        for key in self.type_tree.subtypes:
            item = getattr(self, key)[index]
            item_type = self.type_tree.subtypes[key].type
            # if not isinstance(item, item_type):
            #    item = item_type(item)
            setattr(ret, key, item)
        return ret

    def to(self, device):
        ret = Batch(None)
        ret.type_tree = self.type_tree
        ret.batch_size =  self.batch_size
        ret.store = { key: val.to(device) for key, val in self.store.items() }
        return ret

@default_init
class BatchIter:
    batch_td: "BatchTD"
    index: int

    def __next__(self):
        if self.index != 0:
            raise StopIteration()
        self.index += 1
        return self.batch_td[0]

class BatchTD(ClassTD):

    def __init__(self, type_data: ClassTD, batch_size: Union[int, ShapeVar]):
        subtypes = {}
        for name, subtype in type_data.subtypes.items():
            subtypes[name] = make_batch_td(subtype, batch_size)
        super(BatchTD, self).__init__(Batch[type_data.runtime_type], **subtypes)

    def __getitem__(self, index):
        runtime_type = self.runtime_type.__args__[0]
        subtypes = {}
        for name, subtype in self.subtypes.items():
            subtypes[name] = subtype[index]
        return ClassTD(runtime_type, **subtypes)

    def __iter__(self):
        return BatchIter(self, 0)

def make_batch_td(type_data: TypeData, 
                  batch_size: Union[int, ShapeVar] = ShapeVar('B')) -> TypeData:
    if isinstance(type_data, tuple):
        return tuple(make_batch_td(item) for item in type_data)
    elif isinstance(type_data, TensorTD):
        ret = copy(type_data)
        ret.shape = tuple([batch_size] + list(type_data.shape))
        return ret
    else:
        try:
            return type_data.runtime_type.get_batch_td_type()(type_data, batch_size)
        except NotImplementedError:
            return BatchTD(type_data, batch_size)

def make_batch(items):
    assert len(items) > 0
    first = items[0]
    try:
        return type(first).get_batch_type()(items)
    except NotImplementedError:
        return Batch(items)
            
def collate(batch: Any) -> Any:
    example = batch[0]
    if isinstance(example, Batchable):
        return make_batch(batch)
    elif isinstance(example, tuple) or isinstance(example, list):
        ret = []
        for i, item in enumerate(example):
            all_items = [ b[i] for b in batch]
            ret.append(collate(all_items))
        return type(example)(ret)
    elif isinstance(example, dict):
        raise NotImplementedError()
    else:
        return default_collate(batch)

class DataLoader(torch.utils.data.DataLoader):
    
    def __init__(self, dataset, batch_size=1, shuffle=False, **kwargs):
        super(DataLoader, self).__init__(
            dataset, batch_size, shuffle,
            collate_fn=collate, **kwargs
        )

    def get_type_data(self) -> TypeData:
        """ If the dataset impliments get_type_data, this will return the
        batchfied type data """
        return make_batch_td(self.dataset.get_type_data())

if __name__ == "__main__":

    class Test(Batchable):
        t1: torch.Tensor
        t2: list

    class Test2(Batchable):
        item1: Test
        item2: Test

    class Test3(Batchable):
        fred: Test2
        george: Test

    test = Test(torch.tensor([1,2,3]), t2=[1,2,3])
    test2 = Test2(test, test)
    test3 = Test3(test2, test)
    batch = Batch([test, test])
    batch3 = Batch([test3, test3, test3])
    print(batch3.george.t1)
    



import torch
from torch import nn
from typing import Type, Any, List, Tuple

from .type_data import TypeData, TensorTD
from .meta_utils import default_init, get_any_arg, contains_type, recursive_map, recursive_zip
from .comp_node import CompNode

def call_with_type_data(func, type_func, module, *args, **kwargs):
    """ If args contain type data, use type func, else use func.
    If comp node data, return comp node with module (which could be) """
    # arg = get_any_arg(*args, **kwargs)
    # todo: better way to determine if we're being called with a type or not?
    if contains_type(args, TypeData) or contains_type(kwargs, TypeData):
        return type_func(*args, **kwargs)
    elif contains_type(args, CompNode) or contains_type(kwargs, CompNode):
        def to_type(node):
            return node.out_type_data # if isinstance(node, TypeData) else node
        type_args = recursive_map(to_type, args)
        type_kwargs = recursive_map(to_type, kwargs)
        td = type_func(*type_args, **type_kwargs)
        return CompNode(
            args=args,
            kwargs=kwargs,
            op=module,
            out_type_data=td
        )
    else:
        return func(*args, **kwargs)
    
class TypedModule(nn.Module):

    def __init__(self):
        super(TypedModule, self).__init__()

    def get_shape(self, *args, **kwargs):
        raise NotImplementedError

    def get_shapes(self, *args, **kwargs):
        raise NotImplementedError
        
    def get_type_data(self, *args, **kwargs):
        try:
            return TensorTD(self.get_shape(*args, **kwargs))
        except NotImplementedError:
            return [ TensorTD(shape) for shape in self.get_shapes(*args, **kwargs) ]
        except NotImplementedError:
            raise
        
    def __call__(self, *args, **kwargs):
        return call_with_type_data(super().__call__, self.get_type_data, self, *args, **kwargs)

class WrapperModule(TypedModule):
    submodule: nn.Module

    def __init__(self, submodule: nn.Module):
        super(WrapperModule, self).__init__()
        self.submodule = submodule
    
    def forward(self, *args, **kwargs):
        return self.submodule(*args, **kwargs)

class Linear(WrapperModule):

    def __init__(self, in_feats, out_feats, *args, **kwargs):
        super(Linear, self).__init__(nn.Linear(in_feats, out_feats, *args, **kwargs))
        self.in_feats = in_feats
        self.out_feats = out_feats

    def get_shape(self, tdata):
        batch, in_feats = tdata.shape
        assert in_feats == self.in_feats
        return (batch, self.out_feats)

class Embedding(WrapperModule):

    def __init__(self, num_embed, embed_dim, *args, **kwargs):
        super(Embedding, self).__init__(nn.Embedding(num_embed, embed_dim, *args, **kwargs))
        self.embed_dim = embed_dim

    def get_shape(self, tdata):
        return tuple(list(tdata.shape) + [self.embed_dim])


class MultiheadAttention(WrapperModule):

    def get_dims(self,
                 embed_dim,
                 num_heads,
                 dropout=0.0,
                 bias=True,
                 add_bias_kv=False,
                 add_zero_attn=False,
                 kdim=None,
                 vdim=None, 
                 batch_first=False, 
                 device=None, 
                 dtype=None):
        kdim = embed_dim if kdim is None else kdim
        vdim = embed_dim if vdim is None else vdim
        return embed_dim, kdim, vdim

    def __init__(self, *args, **kwargs):
        super(MultiheadAttention, self).__init__(nn.MultiheadAttention(*args, **kwargs))
        embed_dim, kdim, vdim = self.get_dims(*args, **kwargs)
        self.embed_dim = embed_dim
        self.kdim = kdim
        self.vdim = vdim

    def get_type_data(self, qt, kt, vt):
        # for now, only considering unbatched case all tensors shape (L, E)
        L, Eq = qt.shape
        S, Ek = kt.shape
        shape1 = (L, self.embed_dim)
        shape2 = (L, S)
        return TensorTD(shape1), TensorTD(shape2)

class Activation(WrapperModule):

    def __init__(self, act):
        super(Activation, self).__init__(act)

    def get_type_data(self, tdata):
        return tdata

class ReLU(Activation):

    def __init__(self, *args, **kwargs):
        super(ReLU, self).__init__(nn.ReLU(*args, **kwargs))

class BatchNorm1d(Activation):

    def __init__(self, *args, **kwargs):
        super(BatchNorm1d, self).__init__(nn.BatchNorm1d(*args, **kwargs))

def wraps_function(f):
    """ Wraps a get_type_data function with f so the resulting
    function can take either type data or tensors """
    def wrapper(td_f):
        def inner(*args, **kwargs):
            return call_with_type_data(f, td_f, f, *args, **kwargs)
        return inner
    return wrapper
            
@wraps_function(torch.cat)
def cat(in_types: Tuple[TensorTD, ...], axis: int) -> TensorTD:
    assert len(in_types) > 0
    out_shape = None
    for td in in_types:
        in_shape = td.shape
        if out_shape is None:
            out_shape = list(in_shape)
        else:
            assert len(out_shape) == len(in_shape)
            out_shape[axis] += in_shape[axis]
    return TensorTD(tuple(out_shape), dtype=in_types[0].dtype)


class Model(TypedModule):

    in_nodes: Any # can be any container containing Inputs
    out_nodes: Any # can be any container containing CompNodes

    def __init__(self, in_nodes, out_nodes):
        super(Model, self).__init__()
        self.in_nodes = in_nodes
        self.out_nodes = out_nodes

        # do bfs to determine which submodules we need to register
        self.submodules = nn.ModuleList()
        to_explore = []
        recursive_map(lambda node: to_explore.append(node), out_nodes)
        while len(to_explore) > 0:
            node = to_explore.pop()
            if isinstance(node.op, nn.Module):
                self.submodules.append(node.op)
            recursive_map(lambda parent: to_explore.append(parent), node.args)
            recursive_map(lambda parent: to_explore.append(parent), node.kwargs)

    def forward(self, inputs):
        for node, input in recursive_zip(self.in_nodes, inputs):
            node.set_value(input)
        return recursive_map(lambda node: node.execute(), self.out_nodes)

    def get_type_data(self, in_td, out_td):
        return self.out_type_data

    @property
    def out_type_data(self):
        return recursive_map(lambda node: node.out_type_data, self.out_nodes)


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()
    
setuptools.setup(
    name="terrace",
    version="0.0.1",
    author="Michael Brocidiacono",
    author_email="",
    description="high level PyTorch utils",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mixarcid/terrace",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=['torch>=1.6.0']
)



from typing import *

from .typed_module import TypedModule, Linear, Embedding, MultiheadAttention, BatchNorm1d
from .comp_node import CompNode

class MakeModule:
    """ base class of an object that will generate modules
    given the type data of previous modules in the comp graph """

    def get_module(*args, **kwargs) -> TypedModule:
        raise NotImplementedError()

    # todo: these should input _types_, not compnodes directly
    def __call__(self, *args, **kwargs) -> CompNode:
        module = self.get_module(*args, **kwargs)
        return module(*args, **kwargs)

class MakeWrapper(MakeModule):

    args: List[Any]
    kwargs: Dict[str, Any]

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

class MakeLinear(MakeWrapper):

    def get_module(self, in_node: CompNode) -> Linear:
        return Linear(in_node.shape[-1], *self.args, **self.kwargs)

class MakeBatchNorm1d(MakeWrapper):

    def get_module(self, in_node: CompNode) -> BatchNorm1d:
        return BatchNorm1d(in_node.shape[-1], *self.args, **self.kwargs)

class MakeEmbedding(MakeWrapper):

    def get_module(self, in_node: CompNode) -> Embedding:
        embed_dim = in_node.out_type_data.max_value
        return Embedding(embed_dim, *self.args, **self.kwargs)

class MakeMultiheadAttention(MakeWrapper):

    def get_module(self, qn: CompNode, kn: CompNode, vn: CompNode) -> MultiheadAttention:
        embed_dim = qn.shape[-1]
        self.kwargs["kdim"] = kn.shape[-1]
        self.kwargs["vdim"] = vn.shape[-1]
        return MultiheadAttention(embed_dim, *self.args, **self.kwargs)


import torch
from torch import nn
from typing import Callable, Any, List, Dict

from .type_data import TypeData, TensorTD
from .meta_utils import default_init, recursive_map

@default_init
class CompNode:
    """ Node in the computational graph. Is defined by in nodes,
    out type data, and an operation """

    args: List[Any] # can be any container containing nodes
    kwargs: Dict[str, Any]
    op: Callable
    out_type_data: Any # can be any container containing type data

    def execute(self):
        """ Recursively excutes previous nodes to get the final value """
        exec_args = recursive_map(lambda node: node.execute(), self.args)
        exec_kwargs = recursive_map(lambda node: node.execute(), self.kwargs)
        return self.op(*exec_args, **exec_kwargs)

    @property
    def shape(self):
        return self.out_type_data.shape

    def __getitem__(self, idx):
        op = lambda node: node[idx]
        return CompNode([self], {}, op, self.out_type_data[idx])
    
class Input(CompNode):

    value: Any
    has_set_value: bool

    def __init__(self, type_data):
        super().__init__([], {}, lambda x: x, type_data)
        self.value = None
        self.has_set_value = False

    def set_value(self, value):
        self.value = value
        self.has_set_value = True

    def execute(self):
        assert self.has_set_value
        return self.value



import torch
from copy import deepcopy
from typing import Type, List, Tuple, Union, Dict, Type, Optional
from collections import defaultdict
from sympy import Symbol, Expr, Integer, symbols

from .meta_utils import default_init, get_type_name

def to_expr(item: Union["ShapeVar", int]) -> Expr:
    if isinstance(item, int):
        return Integer(item)
    else:
        return item.expr

class ShapeVar:
    """ If we don't know the dimension of something at graph time,
    put it in a shape var (e.g. batch) """

    expr: Expr

    def __init__(self, expr: Union[Expr, str]):
        if isinstance(expr, str):
            self.expr = symbols(expr, integer=True)
        else:
            self.expr = expr

    # for everything in this library, get_name is used when pretty-printing
    # or exporting models, to pdfs, __repr__ is used for regular printing
    def get_name(self) -> str:
        return str(self.expr)
    
    def __repr__(self) -> str:
        return f"ShapeVar[{self.get_name()}]"

    def maybe_int(self) -> Union["ShapeVar", int]:
        """ If the expression evals to an int, returns an int """
        try:
            return int(self.get_name())
        except ValueError:
            return self

    def __eq__(self, other: Union["ShapeVar", int]):
        return to_expr(other).equals(self.expr).maybe_int()

# somewhat cursed way of generating all the operators
def set_operator(method):
    def op(self, other):
        return ShapeVar(getattr(self.expr, method)(to_expr(other))).maybe_int()
    setattr(ShapeVar, method, op)
    
for method in [ "__add__", "__sub__", "__mul__", "__floordiv__", "__radd__", "__rsub__", "__rmul__", "__rfloordiv__" ]:
    set_operator(method)
    
@default_init
class TypeData:
    """ tabulates all the stuff we need to know about some future
    variable at graph creation time. E.g. the shape of a tensor """

    # the type the variable will be at runtime
    runtime_type: Type

    def __repr__(self) -> str:
        return f"TypeData[{self.get_name()}]"

    def get_name(self) -> str:
        runtime_type_name = self.get_type_name()
        try:
            return f"{runtime_type_name}[{self.get_params_repr()}]"
        except NotImplementedError:
            return runtime_type_name

    def get_type_name(self) -> str:
        return get_type_name(self.runtime_type)

    def get_params_repr(self) -> str:
        raise NotImplementedError

ShapeType = Tuple[Union[int, ShapeVar], ...]

def normalize_idx(sz, idx):
    return idx if idx >= 0 else sz + idx

class TensorTD(TypeData):
    """ Type data for tensor. This is what will be mostly used """
    shape: ShapeType
    dtype: torch.dtype
    # max_values are used to define the max values taken by long tensors
    # for classification. For everything else will be None
    # assume max values apply to the last axis of the tensor
    max_values: Optional[List[int]]

    def __init__(self, shape: ShapeType, 
                 dtype: Type = torch.float32, 
                 max_values: Optional[List[int]] = None):
        super().__init__(torch.Tensor)
        if max_values is not None and isinstance(shape[-1], int):
            assert len(max_values) == shape[-1]
        self.shape = shape
        self.dtype = dtype
        self.max_values = max_values

    def get_params_repr(self) -> str:
        ret = []
        for item in self.shape:
            if isinstance(item, ShapeVar):
                s = item.get_name()
            else:
                s = repr(item)
            ret.append(s)
        ret = "(" + ", ".join(ret) + f")"
        # only print dtype if it ain't float
        if self.dtype != torch.float32:
            ret += f", dtype={str(self.dtype).split('.')[-1]}"
        if self.max_values is not None:
            ret += f", max_values={self.max_values}"
        return ret

    def __len__(self):
        return 1

    def __getitem__(self, idx) -> "TensorTD":
        if not isinstance(idx, tuple):
            idx = (idx,)
        ret_shape = []
        ret_max_values = self.max_values
        for axis, (i, sz) in enumerate(zip(idx, self.shape)):
            if axis == len(self.shape) - 1 and self.max_values is not None:
                ret_max_values = self.max_values[i]
                if isinstance(i, int):
                    ret_max_values = [ ret_max_values ]

            if isinstance(i, int):
                continue
            if isinstance(i, slice):
                if i.step is not None:
                    raise NotImplementedError()
                start = 0 if i.start is None else normalize_idx(sz, i.start)
                stop = sz if i.stop is None else normalize_idx(sz, i.stop)
                if isinstance(start, int) and isinstance(stop, int):
                    assert stop >= start
                    assert stop <= sz
                ret_shape.append(stop-start)
        ret_shape += list(self.shape[len(idx):])
        ret = deepcopy(self)
        ret.shape = tuple(ret_shape)
        ret.max_values = ret_max_values
        return ret

    @property
    def max_value(self):
        assert self.max_values is not None and len(self.max_values) == 1
        return self.max_values[0]

class ClassTD(TypeData):
    """ Type data for custom data types """

    subtypes: Dict[str, TypeData]

    def __init__(self, runtime_type, **kwargs):
        super().__init__(runtime_type)
        self.subtypes = {}
        for key, val in kwargs.items():
            self.subtypes[key] = val

    def __getattr__(self, key: str) -> TypeData:
        if key in [ "runtime_type", "subtypes" ]:
            raise AttributeError
        else:
            try:
                return self.subtypes[key]
            except KeyError:
                raise AttributeError
        
    def __setattr__(self, key: str, val: TypeData):
        if key in [ "runtime_type", "subtypes" ]:
            super().__setattr__(key, val)
        else:
            self.subtypes[key] = val

    def get_params_repr(self) -> str:
        ret = []
        for key, val in self.subtypes.items():
            ret.append(f"{key}={val.get_name()}")
        return ", ".join(ret)
